use std::{
    collections::HashMap,
    ops::{Deref, DerefMut},
    sync::Arc,
    time::Duration,
};

use jabber::{connection::Tls, jabber_stream::bound_stream::BoundJabberReader};
use sqlx::SqlitePool;
use stanza::client::Stanza;
use tokio::{
    sync::{mpsc, oneshot, Mutex},
    task::{JoinHandle, JoinSet},
};

use crate::{error::Error, UpdateMessage};

use super::{
    write::{WriteHandle, WriteMessage},
    SupervisorCommand,
};

pub struct Read {
    // TODO: place iq hashmap here
    control_receiver: mpsc::Receiver<ReadControl>,
    stream: BoundJabberReader<Tls>,
    on_crash: oneshot::Sender<(
        SqlitePool,
        mpsc::Sender<UpdateMessage>,
        JoinSet<()>,
        mpsc::Sender<SupervisorCommand>,
        WriteHandle,
        Arc<Mutex<HashMap<String, oneshot::Sender<Result<Stanza, Error>>>>>,
    )>,
    db: SqlitePool,
    update_sender: mpsc::Sender<UpdateMessage>,
    supervisor_control: mpsc::Sender<SupervisorCommand>,
    write_handle: WriteHandle,
    tasks: JoinSet<()>,
    disconnecting: bool,
    disconnect_timedout: oneshot::Receiver<()>,
    // TODO: use proper stanza ids
    pending_iqs: Arc<Mutex<HashMap<String, oneshot::Sender<Result<Stanza, Error>>>>>,
}

impl Read {
    fn new(
        control_receiver: mpsc::Receiver<ReadControl>,
        stream: BoundJabberReader<Tls>,
        on_crash: oneshot::Sender<(
            SqlitePool,
            mpsc::Sender<UpdateMessage>,
            JoinSet<()>,
            mpsc::Sender<SupervisorCommand>,
            WriteHandle,
            Arc<Mutex<HashMap<String, oneshot::Sender<Result<Stanza, Error>>>>>,
        )>,
        db: SqlitePool,
        update_sender: mpsc::Sender<UpdateMessage>,
        // jabber server must be able to both terminate the connection from error, and ask for data from the client (such as supported XEPs)
        supervisor_control: mpsc::Sender<SupervisorCommand>,
        write_handle: WriteHandle,
        tasks: JoinSet<()>,
        pending_iqs: Arc<Mutex<HashMap<String, oneshot::Sender<Result<Stanza, Error>>>>>,
    ) -> Self {
        let (send, recv) = oneshot::channel();
        Self {
            control_receiver,
            stream,
            on_crash,
            db,
            update_sender,
            supervisor_control,
            write_handle,
            tasks,
            disconnecting: false,
            disconnect_timedout: recv,
            pending_iqs,
        }
    }

    async fn run(mut self) {
        println!("started read thread");
        // let stanza = self.stream.read::<Stanza>().await;
        // println!("{:?}", stanza);
        loop {
            tokio::select! {
                // if still haven't received the end tag in time, just kill itself
                // TODO: is this okay??? what if notification thread dies?
                Ok(()) = &mut self.disconnect_timedout => {
                    println!("disconnect_timedout");
                    break;
                }
                Some(msg) = self.control_receiver.recv() => {
                    match msg {
                        // when disconnect received,
                        ReadControl::Disconnect => {
                            let (send, recv) = oneshot::channel();
                            self.disconnect_timedout = recv;
                            self.disconnecting = true;
                            tokio::spawn(async {
                                tokio::time::sleep(Duration::from_secs(10)).await;
                                let _ = send.send(());
                            })
                        },
                        ReadControl::Abort(sender) => {
                            let _ = sender.send((self.db, self.update_sender, self.tasks, self.supervisor_control, self.write_handle, self.pending_iqs.clone()));
                            break;
                        },
                    };
                },
                s = self.stream.read::<Stanza>() => {
                    println!("read stanza");
                    match s {
                        Ok(s) => {
                            self.tasks.spawn(handle_stanza(s, self.update_sender.clone(), self.db.clone(), self.supervisor_control.clone(), self.write_handle.clone()));
                        },
                        Err(e) => {
                            println!("error: {:?}", e);
                            // TODO: NEXT write the correct error stanza depending on error, decide whether to reconnect or properly disconnect, depending on if disconnecting is true
                            // match e {
                            //     peanuts::Error::ReadError(error) => todo!(),
                            //     peanuts::Error::Utf8Error(utf8_error) => todo!(),
                            //     peanuts::Error::ParseError(_) => todo!(),
                            //     peanuts::Error::EntityProcessError(_) => todo!(),
                            //     peanuts::Error::InvalidCharRef(_) => todo!(),
                            //     peanuts::Error::DuplicateNameSpaceDeclaration(namespace_declaration) => todo!(),
                            //     peanuts::Error::DuplicateAttribute(_) => todo!(),
                            //     peanuts::Error::UnqualifiedNamespace(_) => todo!(),
                            //     peanuts::Error::MismatchedEndTag(name, name1) => todo!(),
                            //     peanuts::Error::NotInElement(_) => todo!(),
                            //     peanuts::Error::ExtraData(_) => todo!(),
                            //     peanuts::Error::UndeclaredNamespace(_) => todo!(),
                            //     peanuts::Error::IncorrectName(name) => todo!(),
                            //     peanuts::Error::DeserializeError(_) => todo!(),
                            //     peanuts::Error::Deserialize(deserialize_error) => todo!(),
                            //     peanuts::Error::RootElementEnded => todo!(),
                            // }
                            // TODO: make sure this only happens when an end tag is received
                            if self.disconnecting == true {
                                break;
                            } else {
                                // AAAAAAAAAAAAAAAAAAAAA i should really just have this stored in the supervisor and not gaf bout passing these references around
                                let _ = self.on_crash.send((self.db, self.update_sender, self.tasks, self.supervisor_control, self.write_handle, self.pending_iqs.clone()));
                            }
                            break;
                        },
                    }
                },
                else => break
            }
        }
        println!("stopping read thread");
        // when it aborts, must clear iq map no matter what
        let mut iqs = self.pending_iqs.lock().await;
        for (_id, sender) in iqs.drain() {
            let _ = sender.send(Err(Error::LostConnection));
        }
    }
}

// what do stanza processes do?
// - update ui
// - access database
// - disconnect proper, reconnect
// - respond to server requests
async fn handle_stanza(
    stanza: Stanza,
    update_sender: mpsc::Sender<UpdateMessage>,
    db: SqlitePool,
    supervisor_control: mpsc::Sender<SupervisorCommand>,
    write_handle: WriteHandle,
) {
    println!("{:?}", stanza)
}

pub enum ReadControl {
    Disconnect,
    Abort(
        oneshot::Sender<(
            SqlitePool,
            mpsc::Sender<UpdateMessage>,
            JoinSet<()>,
            mpsc::Sender<SupervisorCommand>,
            WriteHandle,
            Arc<Mutex<HashMap<String, oneshot::Sender<Result<Stanza, Error>>>>>,
        )>,
    ),
}

pub struct ReadControlHandle {
    sender: mpsc::Sender<ReadControl>,
    pub(crate) handle: JoinHandle<()>,
}

impl Deref for ReadControlHandle {
    type Target = mpsc::Sender<ReadControl>;

    fn deref(&self) -> &Self::Target {
        &self.sender
    }
}

impl DerefMut for ReadControlHandle {
    fn deref_mut(&mut self) -> &mut Self::Target {
        &mut self.sender
    }
}

impl ReadControlHandle {
    pub fn new(
        stream: BoundJabberReader<Tls>,
        on_crash: oneshot::Sender<(
            SqlitePool,
            mpsc::Sender<UpdateMessage>,
            JoinSet<()>,
            mpsc::Sender<SupervisorCommand>,
            WriteHandle,
            Arc<Mutex<HashMap<String, oneshot::Sender<Result<Stanza, Error>>>>>,
        )>,
        db: SqlitePool,
        sender: mpsc::Sender<UpdateMessage>,
        supervisor_control: mpsc::Sender<SupervisorCommand>,
        jabber_write: WriteHandle,
        pending_iqs: Arc<Mutex<HashMap<String, oneshot::Sender<Result<Stanza, Error>>>>>,
    ) -> Self {
        let (control_sender, control_receiver) = mpsc::channel(20);

        let actor = Read::new(
            control_receiver,
            stream,
            on_crash,
            db,
            sender,
            supervisor_control,
            jabber_write,
            JoinSet::new(),
            pending_iqs,
        );
        let handle = tokio::spawn(async move { actor.run().await });

        Self {
            sender: control_sender,
            handle,
        }
    }

    pub fn reconnect(
        stream: BoundJabberReader<Tls>,
        on_crash: oneshot::Sender<(
            SqlitePool,
            mpsc::Sender<UpdateMessage>,
            JoinSet<()>,
            mpsc::Sender<SupervisorCommand>,
            WriteHandle,
            Arc<Mutex<HashMap<String, oneshot::Sender<Result<Stanza, Error>>>>>,
        )>,
        db: SqlitePool,
        sender: mpsc::Sender<UpdateMessage>,
        supervisor_control: mpsc::Sender<SupervisorCommand>,
        jabber_write: WriteHandle,
        tasks: JoinSet<()>,
        pending_iqs: Arc<Mutex<HashMap<String, oneshot::Sender<Result<Stanza, Error>>>>>,
    ) -> Self {
        let (control_sender, control_receiver) = mpsc::channel(20);

        let actor = Read::new(
            control_receiver,
            stream,
            on_crash,
            db,
            sender,
            supervisor_control,
            jabber_write,
            tasks,
            pending_iqs,
        );
        let handle = tokio::spawn(async move { actor.run().await });

        Self {
            sender: control_sender,
            handle,
        }
    }
}
